import csv
import pickle

from .multiple_objective import MultipleObjective


def get_best_trials(multiple_objective: MultipleObjective) -> None:
    """
    Extracts and saves the best trials from the given MultipleObjective's log file.

    This function reads the 'paradance.log' file associated with the provided MultipleObjective object,
    and identifies the best trials based on the comparison of values. These trials, along with their
    results, targets, and weights are then saved to a CSV file named 'paradance_best_trials.csv' in
    the same directory.

    Args:
        multiple_objective (MultipleObjective): An instance of the MultipleObjective class containing the study to be saved.
    """
    ob = multiple_objective
    file_path = f"{ob.full_path}/paradance.log"
    output_path = f"{ob.full_path}/paradance_best_trials.csv"

    last_best_value = None
    last_trial_number = -1
    extracted_data = []

    with open(file_path, "r") as file:
        lines = file.readlines()

        for idx, line in enumerate(lines):
            if "Best is trial" in line:
                current_best_value = float(line.split("with value:")[1].split(" ")[1])
                if last_best_value is None or last_best_value != current_best_value:
                    trial_number = int(line.split("Best is trial")[1].split(" ")[1])
                    if trial_number > last_trial_number:
                        results_line = lines[idx - 3]
                        targets_line = lines[idx - 2]
                        weights_line = lines[idx - 1]

                        results = float(results_line.split("result:")[1].strip())
                        targets_str = [
                            val
                            for val in targets_line.split(":")[1]
                            .strip()
                            .strip("[]\n")
                            .split(",")
                        ]
                        weights_str = [
                            val
                            for val in weights_line.split(":")[1]
                            .strip()
                            .strip("[]\n")
                            .split(", ")
                        ]

                        if len(weights_str) != 1:
                            targets = [float(val) for val in targets_str]
                            weights = [float(val) for val in weights_str]
                            extracted_data.append(
                                (results, trial_number, targets, weights)
                            )

                            last_trial_number = trial_number
                            last_best_value = current_best_value

    with open(output_path, "w", newline="") as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(["Result", "Trial", "Targets", "Weights"])
        for data in extracted_data:
            writer.writerow([data[0], data[1], str(data[2]), str(data[3])])


def save_study(multiple_objective: MultipleObjective) -> None:
    """Save the study results of the given multiple objective optimization.

    Args:
        multiple_objective (MultipleObjective): An instance of the MultipleObjective class containing the study to be saved.
    """
    ob = multiple_objective
    get_best_trials(ob)
    ob.study.trials_dataframe().to_csv(f"{ob.full_path}/paradance_full_trials.csv")
    with open(f"{ob.full_path}/study.pkl", "wb") as f:
        pickle.dump(ob.study, f)



from typing import TYPE_CHECKING, List

from sklearn.metrics import roc_auc_score

if TYPE_CHECKING:
    from .calculator import Calculator


def calculate_woauc(
    calculator: "Calculator",
    target_column: str,
    weights_for_equation: List,
) -> List[float]:
    """Calculate weighted ordinal user AUC.

    :param weights_for_equation: weights for equation
    :param target_column: score column
    """
    woauc_indices = calculator.woauc_dict[target_column]
    woauc = []
    sampler = calculator.samplers[target_column]
    for k, _ in sampler.boundary_dict.items():
        paritial_auc = float(
            roc_auc_score(
                calculator.df.loc[woauc_indices][f"{target_column}_lt_{k}"],
                calculator.df.loc[woauc_indices]["overall_score"],
            )
        )
        woauc.append(paritial_auc)
    return woauc



from functools import partialmethod
from typing import List, Optional

from optuna.trial import Trial

from ..evaluation.calculator import Calculator
from .base import BaseObjective
from .construct_weights import construct_weights
from .evaluate_targets import evaluate_targets


class MultipleObjective(BaseObjective):
    """
    This class provides methods to optimize the portfolio objective.
    """

    construct_weights = partialmethod(construct_weights)

    def __init__(
        self,
        calculator: Calculator,
        direction: str,
        weights_num: int,
        formula: str,
        power: bool = False,
        first_order: bool = False,
        first_order_lower_bound: float = 1e-3,
        first_order_upper_bound: float = 1e6,
        power_lower_bound: float = 0,
        power_upper_bound: float = 1,
        dirichlet: bool = True,
        study_name: Optional[str] = None,
        study_path: Optional[str] = None,
    ) -> None:
        """
        Initialize with direction, weights_num, formula and dirichlet.

        Args:
            direction (str ["minimize", "maximize"]): direction to optimize.
            weights_num (int): numbers of weights to search.
            formula (str): formula of targets to calculate the objective.
            dirichlet (bool, optional): Use dirichlet distribution or not. Defaults to True.
        """
        super().__init__(direction=direction, formula=formula)
        self.power = power
        self.formula = formula
        self.first_order = first_order
        self.weights_num = weights_num
        self.target_columns: List[str] = []
        self.evaluator_flags: List[str] = []
        self.groupbys: List[Optional[str]] = []
        self.calculator: Calculator = calculator
        self.power_lower_bound = power_lower_bound
        self.power_upper_bound = power_upper_bound
        self.hyperparameters: List[Optional[float]] = []
        self.evaluator_propertys: List[Optional[str]] = []
        self.first_order_lower_bound = first_order_lower_bound
        self.first_order_upper_bound = first_order_upper_bound
        if self.power_lower_bound < 0:
            self.dirichlet = False
        else:
            self.dirichlet = dirichlet

    def add_evaluator(
        self,
        flag: str,
        target_column: str,
        hyperparameter: Optional[float] = None,
        evaluator_property: Optional[str] = None,
        groupby: Optional[str] = None,
    ) -> None:
        """Add calculators to the objective.

        Args:
            calculator (Calculator): calculator building blocks.
            flag (str ["wuauc", "portfolio", "logmse", ..., ect.]): type of calculator.
            target_column (str): target column to calculate.
        """
        self.evaluator_flags.append(flag)
        self.target_columns.append(target_column)
        if hyperparameter is not None:
            self.hyperparameters.append(hyperparameter)
        else:
            self.hyperparameters.append(None)
        if groupby is not None:
            self.groupbys.append(groupby)
        else:
            self.groupbys.append(None)
        if evaluator_property is not None:
            self.evaluator_propertys.append(evaluator_property)
        else:
            self.evaluator_propertys.append(None)

    def objective(
        self,
        trial: Trial,
    ) -> float:
        """Objective function for optuna.

        Args:
            trial (Trial): optuna trial.

        Returns:
            float: objective value.
        """
        weights = construct_weights(self, trial)
        targets = evaluate_targets(
            calculator=self.calculator,
            evaluator_flags=self.evaluator_flags,
            hyperparameters=self.hyperparameters,
            evaluator_propertys=self.evaluator_propertys,
            groupbys=self.groupbys,
            target_columns=self.target_columns,
            weights=weights,
        )
        local_vars = {"targets": targets, "sum": sum, "max": max, "min": min}
        result = float(eval(self.formula, {"__builtins__": None}, local_vars))
        if self.logger:
            self.logger.info(f"Trial {trial.number} finished with result: {result}")
            self.logger.info(f"targets: {targets}")
            self.logger.info(f"weights: {weights}")
        return result



import logging
from abc import ABCMeta, abstractmethod
from typing import Optional

import optuna

from .set_path import ensure_study_directory


class BaseObjective(metaclass=ABCMeta):
    """
    This class provides methods to optimize obejective.
    """

    def __init__(
        self,
        direction: str,
        formula: str,
        fist_order: bool = False,
        dirichlet: bool = True,
        study_name: Optional[str] = None,
        study_path: Optional[str] = None,
    ) -> None:
        self.full_path = ensure_study_directory(study_path, study_name)
        storage_path = f"sqlite:///{self.full_path}/paradance_storage.db"
        self.study = optuna.create_study(
            direction=direction,
            study_name=study_name,
            storage=storage_path,
            load_if_exists=True,
        )

    def build_logger(self, process_id: Optional[int] = None) -> None:
        """Build logger for paradance.

        Args:
            log_file (str): log file path.
            process_id (int, optional): ID of the process. Used to create unique log filenames in parallel execution.
        """

        if process_id is not None:
            log_filename = f"{self.full_path}/paradance_{process_id}.log"
        else:
            log_filename = f"{self.full_path}/paradance.log"

        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(logging.INFO)
        self.logger = optuna.logging.get_logger(
            f"paradance_{process_id}" if process_id else "optuna"
        )
        self.logger.addHandler(file_handler)

    @abstractmethod
    def objective(self, trial: optuna.trial.Trial) -> float:
        """Objective function to optimize."""
        raise NotImplementedError

    def optimize(self, n_trials: int) -> None:
        """Optimize the objective."""
        self.build_logger()
        self.study.optimize(self.objective, n_trials=n_trials)



from typing import TYPE_CHECKING, List

import optuna

if TYPE_CHECKING:
    from .multiple_objective import MultipleObjective


def construct_weights(ob: "MultipleObjective", trial: optuna.Trial) -> List[float]:
    """
    Construct a list of weights based on the attributes of the given MultipleObjective instance
    and the current optuna trial.

    Args:
        ob (MultipleObjective): An instance of the MultipleObjective class, encapsulating various
                                attributes like 'power', 'first_order', 'weights_num', and so forth,
                                which guide the construction of the weight values.
        trial (optuna.Trial): The current optuna trial from which to suggest values for the weights
                              based on the properties of the given MultipleObjective instance.

    Returns:
        List[float]: A list of weight values constructed based on the given MultipleObjective
                     instance and the current optuna trial.
    """
    power_weights: List[float] = []
    first_order_weights: List[float] = []

    if ob.power:
        if ob.dirichlet:
            for i in range(ob.weights_num - 1):
                power_weights.append(
                    trial.suggest_float(
                        f"w_po_{i+1}", 0, max((1 - sum(power_weights), 0.1))
                    )
                )
            power_weights.append(1 - sum(power_weights))
        else:
            for i in range(ob.weights_num):
                power_weights.append(
                    trial.suggest_float(
                        f"w{i+1}", ob.power_lower_bound, ob.power_upper_bound
                    )
                )

    if ob.first_order:
        if ob.first_order_lower_bound < 0:
            log = False
        else:
            log = True
        for i in range(ob.weights_num):
            first_order_weights.append(
                trial.suggest_float(
                    f"w_fo_{i+1}",
                    ob.first_order_lower_bound,
                    ob.first_order_upper_bound,
                    log=log,
                )
            )

    weights: List[float] = []
    if not (ob.first_order):
        weights = power_weights
    elif ob.calculator.equation_type == "product" and ob.power:
        weights = power_weights + first_order_weights
    elif ob.calculator.equation_type == "sum":
        weights = first_order_weights
    elif ob.calculator.equation_type == "free_style":
        weights = first_order_weights
    ob.calculator.get_overall_score(
        weights_for_equation=weights,
    )
    return weights



import os

import pandas as pd

from .base import BaseDataLoader


class ExcelLoader(BaseDataLoader):
    "ExcelLoader class for loading excel files"

    def __init__(self, file_path: str, **kwargs: str):
        super().__init__(file_path, **kwargs)
        self.column_name_spliting()

    def load_data(self) -> pd.DataFrame:
        """Load data from excel file."""
        if self.file_name is not None:
            file_url = os.path.join(self.file_path, self.file_name) + ".xlsx"
            return pd.read_excel(file_url)
        else:
            files = os.listdir(self.file_path)
            df_list = []
            for file in files:
                if file.endswith(self.file_type):
                    file_url = os.path.join(self.file_path, file)
                    df_list.append(pd.read_excel(file_url))
            return pd.concat(df_list)
